package IPhotoXML;
use strict;
use warnings;
use Data::Dumper;
use Encode;

sub abs2timet { return shift() + 978303600 }; # abs base is 1.1.2001

sub load {
	my ($self,$file,%args) = @_;
	$self = bless \%args, ( ref($self) || $self );

	open( my $fh,'<:utf8', $file ) || die $!;
	$_ = do { local $/; <$fh> };
	s{\A.*?<plist[^>]*>\s*}{}s;
	s{\s*</plist>.*\Z}{}s;

	my (@key,$val,$tree,$root,@stack);
	my $set = sub {
		my $v = shift;
		if ( ! $root ) {
			$root = $tree = $v
		} elsif ( @key ) {
			$tree->{$key[-1]} = $v;
			pop @key;
		} else {
			push @$tree,$v;
		}
		if ( ref($v)) {
			#warn "Enter $v\n";
			push @stack,( $tree = $v );
		}
	};

	while ( m{\G(?:<(\w+)(?:\s[^>]*)?/?>|</(\w+)>|([^<]+))}sgc ) {
		if ( defined $1 ) {
			# Starttag
			#warn "start=$1\n";
			$val = undef; 
			my $v;
			if ( $1 eq 'dict' ) {
				$v = {};
			} elsif ( $1 eq 'array' ) {
				$v = [];
			} elsif ( $1 eq 'true' ) {
				$v = 1;
			} elsif ( $1 eq 'false' ) {
				$v = 0;
			} elsif ( $1 eq 'key' ) {
				push @key, undef;
			}
			$set->($v) if defined $v;

		} elsif ( defined $2 ) {
			#warn "end=$2\n";
			if ( $2 eq 'key' ) {
				$key[-1] = $val;
				$val = undef;
			} elsif ( $2 eq 'integer' or $2 eq 'real' or $2 eq 'string' ) {
				$set->( $val);
			} elsif ( $2 eq 'array' or $2 eq 'dict' ) {
				#warn "leave $stack[-1] -> $stack[-2]\n";
				pop @stack;
				$tree = $stack[-1];
			}
		} else {
			#warn "text=$3\n";
			$val = '' if ! defined $val;
			$val .= $3;
		}
	}
	m{\G(.{1,40})}gsc;
	die "Unexpected Input++++\n'$1'" if $1;

	my (%files,%albums,%rolls,@keywords);
	$self->{Files} = \%files;
	$self->{Albums} = \%albums;
	$self->{Rolls} = \%rolls;
	$self->{Keywords} = \@keywords;

	my $base = $root->{'Archive Path'} || die;
	$base =~s{/+$}{};
	$self->{Base} ||= $base;

	my $id2kw = $root->{'List of Keywords'} || {};
	@keywords = map { $id2kw->{$_} } sort { $a <=> $b } keys %$id2kw;

	my (%roll2date,%fnames);
	while ( my ($id,$img) = each %{ $root->{'Master Image List'} || die }) {
		my $path = $img->{ImagePath} || die;
		_fixname($path);
		$path =~s{^\Q$base\E/}{} || die "path=$path base=$base";

		# make filename based on Date of image
		my $date = abs2timet($img->{DateAsTimerInterval});
		my @t = localtime($date);
		my $fname;
		my $rand = '';
		my ($ext) = $path =~m{\.(\w+)$} or die $path;
		while(1) {
			$fname = sprintf "%02d%02d%02d_%02d%02d%02d%s.%s",
				$t[5]%100,$t[4]+1,$t[3],$t[2],$t[1],$t[0],$rand,$ext;
			last if ! $fnames{$fname}++;
			$rand = sprintf "_%x",rand(2**32);
		}

		# if caption is the filename ignore it
		my $caption = $img->{Caption};
		( my $rcaption = $caption ) =~s{(\.\w+)?(\s+(Kopie|Copy))?$}{}i;
		$caption = undef if $path =~m{/\Q$rcaption\E(?:_\d+)?(?:\.\w+)?$}i;
		$caption = undef if $caption && $img->{Comment} =~m{^\s*\Q$caption\E\s*$}i;
		my %data = ( 
			Path => $path,
			Name => $fname,
			$caption ? ( Caption => $caption ):(),
		);
		$data{Rating} = $img->{Rating} if $img->{Rating}>0;
		$data{Comment} = $img->{Comment} if $img->{Comment};
		if ( my @kw = grep { $_ } map { $id2kw->{$_} } @{ $img->{Keywords} || [] } ) {
			$data{Keywords} = \@kw
		}
		$files{$id} = \%data;

		# set date of roll to smallest date of image in roll
		my $roll = $img->{Roll};
		$roll2date{$roll} = $date if ! $roll2date{$roll} || $roll2date{$roll}> $date;
	}

	for my $alb (@{ $root->{'List of Albums'} || die }) {
		my $id = $alb->{AlbumId};
		next if $id >= 999000; # ignore special (LastRoll, Library)
		my $type = $alb->{'Album Type'};
		_fixname($alb->{AlbumName});
		if ( $type eq 'Regular' or $type eq 'Folder' ) {
			my @files = map { $files{$_} || die $_ } @{ $alb->{KeyList}};
			$albums{ $id } = {
				Files => \@files,
				Parent => $alb->{Parent},
				Name => $alb->{AlbumName},
			};
		}
	}

	# remove files from folder types which are in the subfolders
	for my $alb (values %albums) {
		if ( !$alb->{Parent} || ! $albums{ $alb->{Parent}} ) {
			delete $alb->{Parent};
			next;
		}
		my %files = map { $_ => 1 } @{$alb->{Files}};
		my $p = $alb;
		while (1) {
			$p = $albums{$p->{Parent} || last} || last;
			if ( my $f = delete $p->{Files} ) {
				@$f = grep { ! $files{$_} } @$f;
				$p->{Files} = $f if @$f;
			}
		}
	}

	for my $roll (@{ $root->{'List of Rolls'} || die }) {
		my $id = $roll->{AlbumId};
		next if $id >= 999000; # ignore special (LastRoll, Library)
		my @files = map { $files{$_} || die $_ } @{ $roll->{KeyList}};
		my @t = localtime( $roll2date{$id} );
		_fixname($roll->{AlbumName});
		$roll->{AlbumName} =~s{^\d{6,}\s+(-\s+)?}{};
		my $name = sprintf "%02d%02d%02d %s", $t[5] % 100,$t[4]+1,$t[3],$roll->{AlbumName};
		$name = sprintf "%02d%02d%02d.%02d%02d %s", $t[5] % 100,$t[4]+1,$t[3],$t[2],$t[1],$roll->{AlbumName}
			if $roll->{$name};
		$rolls{ $name } = \@files;
	}

	return $self;
}

sub _fixname {
	# fix broken UTF-8 stuff in filenames coming from mixing
	# MacOSX and Linux 
	for (@_) {
		s{a\x{308}}{}g; s{o\x{308}}{}g; s{u\x{308}}{}g;
		s{A\x{308}}{}g; s{O\x{308}}{}g; s{U\x{308}}{}g;
	}
}
1;

__END__
######################## Struktur plist iPhoto
- root ist plist
- hashes via dict
   key
   string - string
   integer - integer value
   real - real value
   array - array, welches wiederum zB string oder dict oder array(?)
     drin haben kann
   dict auch mglich?
   </true> - Boolean true
   </false> - Boolean true


im weiteren Notation wie in fw.cfg, d.h. 
  plist.key=value  - dict
  plist.0 = fasel
  plist.0.key = value
  ...


plist.List_of_Albums.#
	.AlbumName = Name des Albums
	.KeyList.# = (string) Id des Bildes
	.AlbumId   = (int) 999XXX scheint special zu sein ???
	.Album_Type = kann zB "Special Roll" sein (bei Last Roll), 'Regular'
	.Parent    = (int) Id des bergeordneten Albums
plist.List_of_Rolls.#
	.AlbumName = Name der Rolle
	.KeyList.# = (string) Id des Bildes
	.Parent    = 999000 (Library)
	.AlbumId   = (int) Nummer des Albums

plist.List_of_Keywords.% = Keyword, Keys sind Nummern
plist.Master_Image_List.%  - Key ist Nummer des Bildes
	MediaType => 'Image'|...
	Rating    => (int) 0...5 ??
	Roll      => (int) Nummer der Roll
	{Date,ModDate,MetaModDate}AsTimerInterval = (real) time_t ?? (wohl kein time_t)
	ImagePath => Pfad
	ThumbPath => ...
	Keywords => #. Liste von Strings mit Nummern des Keywords


*AsTimerInterval sind Sekunden seit 1.1.2001
