use strict;
use warnings;
use IPhotoXML;
use File::Path;
use Data::Dumper;
use Encode;
use Cwd 'realpath';
use Image::ExifTool;
use Getopt::Long qw(:config posix_default bundling);
use File::Copy;

my $DEBUG=0;

my $ibase = "$ENV{HOME}/Pictures/iPhoto Library";
my $abase = "$ENV{HOME}/Pictures/Albums";

my ($rbase,$xml,$copy);
GetOptions(
	'L|iphoto-library-path=s' => \$ibase,
	'R|path-to-rolls=s' => \$rbase,
	'A|path-to-albums=s' => \$abase,
	'X|album-xml=s' => \$xml,
	'c|copy' => \$copy,
	'd|debug'=> \$DEBUG,
	'h|help' => sub {usage()},
);

$xml ||= "$ibase/AlbumData.xml";

# Load Data
my $p = IPhotoXML->load( $xml, Base => $ibase );

# if $copy: stage1 = copy, stage2 = hard-link,
# if ! $copy: stage1 = hard-link, stage2 = symlink
my $stage1_link = do {
	sub {
		my ($finfo,$npath) = @_;
		_fixpath( $npath );
		if ( my $p = $finfo->{CopyPath} ) {
			# everything done already, just hard-link again to new path
			link( $p,$npath );
			return 1;
		}
		my $opath = $finfo->{Path}; # Original path
		$opath = "$ibase/$opath" if $opath !~m{^/};
		_fixpath( $opath );
		if ( $copy ) {
			copy($opath,$npath) || die "copy $opath,$npath: $!";
		} else {
			link($opath,$npath) || die "link $opath,$npath: $!";
		}
		$finfo->{CopyPath} = $npath;

		# do we have an original version?
		if ( $opath =~s{(.*)/Modified/}{$1/Originals/}
			&& -f $opath ) {
			if ( $copy ) { 
				copy( $opath, "$npath.orig" ) &&
					DEBUG( "copied original to $npath.orig" );
			} else {
				link( $opath, "$npath.orig" ) &&
					DEBUG( "linked original to $npath.orig" );
			}
		}

		# add Info to linked/copied image
		if ( $npath =~m{\.jpe?g$}i ) {
			# add keywords, rating, comments and caption into IPTC/Exif
			my %args;
			for( qw( Caption Keywords Comment Rating )) {
				$args{lc($_)} = $finfo->{$_} if defined $finfo->{$_};
			}
			if ( ! %args ) {
				# quick check if maybe "Applemark" comment needs to be removed
				open( my $f,'<',$npath ) || die "cannot open linked image $npath: $!";
				my $t = do { local $/; <$f> };
				DEBUG( "remove applemark comment from $npath" );
				$args{Comment} = undef if $t =~m{AppleMark};
			}
			if ( %args ) {
				set_image_info( $npath, %args ) or 
					warn "failed to set image info on $npath\n"
			}
		}
		return 1;
	}
};

my $stage2_link = sub {
	my ($f,$dst) = @_;
	my $src = $f->{CopyPath};
	if ( ! $src ) {
		$src = $f->{Path};
		$src = "$ibase/$src" if $src !~m{^/};
		_fixpath( $src );
	}
	if ( $copy ) {
		link( $src,$dst ) || die "cannot link $src,$dst: $!";
	} else {
		$src = realpath($src);
		symlink( $src,$dst ) || die "cannot symlink $src,$dst: $!";
	}
};

my @linksub = ( $stage1_link,$stage2_link );

# Create folder for Album
-d $abase or mkdir($abase) or die "mkdir $abase: $!";


# if $rbase is given link images into folder for Rolls
if ( defined $rbase ) {
	my $linksub = shift @linksub;
	-d $rbase or mkdir($rbase) or die "mkdir $rbase: $!"; # create folder
	while ( my ($roll,$files) = each %{ $p->{Rolls}} ) {
		_fixpath( $roll );
		$roll = _mkpath( $rbase,$roll );
		for my $f (@$files) {
			$linksub->( $f,"$roll/$f->{Name}" );
		}
	}
}

my $linksub = shift @linksub;
my $albums = $p->{Albums};
while ( my ($id,$data) = each %$albums ) {
	my $files = $data->{Files} or next;

	# create folder hierarchy for album based on album parents
	my @p = ( $id );
	while (1) {
		my $parent = $albums->{$p[0]}{Parent} or last;
		unshift @p,$parent;
	}
	@p = map { $albums->{$_}{Name} } @p;
	my $base = _mkpath( $abase,@p );

	for my $f (@$files) {
		$linksub->( $f,"$base/$f->{Name}" );
	}
}


sub _mkpath {
	my ($base,@dir) = @_;
	s{[\000/\n\r\t\377]}{}g for(@dir); # UNIX fs allows nearly everything
	_fixpath( $base,@dir );
	my $dir = join( "/",$base,@dir);
	-d $dir or mkpath( $dir ) or die "mkpath $dir: $!";
	return $dir;
}

sub _fixpath {
	# we consider local FS as iso-8859-15 encoded while Apple uses utf-8. 
	# FIXME: should better check locale for it
	$_ = encode( 'iso-8859-15', $_ ) for (@_);
}

# set Exif/IPTC Info for comment, title, keywords and rating
sub set_image_info {
	my ($file,%args) = @_;
	! defined $args{$_} && delete $args{$_} for ( keys %args );
	return if ! %args; 
	delete $args{force};

	my $rating = delete $args{rating};
	my $keywords = delete $args{keywords};
	my $comment = delete $args{comment};
	my $caption = delete $args{caption};

	if ( $rating ) {
		# assign rating to urgncy level same way as digikam 0.9.3
		$rating = int( $rating * -1.4+8.7 );
		$rating = undef if $rating<0;
	}

	my $x = Image::ExifTool->new;
	$x->ExtractInfo($file) || return;
	my $old_rating = $x->GetValue( 'Urgency' );
	my $old_keywords = $x->GetValue( 'Keywords','ValueConv' );
	$old_keywords = [ $old_keywords ] if defined $old_keywords && !ref($old_keywords);
	my $old_comment = $x->GetValue( 'Comment' );
	my $old_caption = $x->GetValue( 'FileName' );
	my $changed = 0;

	if ( $rating ) {
		if ( ! defined $old_rating or $rating != $old_rating ) {
			$x->SetNewValue( 'Urgency', $rating );
			$changed++;
		}
	} elsif ( defined $old_rating ) {
		$x->SetNewValue( 'Urgency' );
		$changed++;
	}

	if ($keywords && @$keywords) {
		if ( ! defined $old_keywords or ! @$old_keywords or 
			join( '\000', sort @$old_keywords ) ne join( '\000',sort @$keywords)) {
			$x->SetNewValue( 'Keywords', $keywords );
			$changed++;
		}
	} elsif ( defined $old_keywords ) {
		$x->SetNewValue( 'Keywords' );
		$changed++;
	}

	if ( defined $comment && $comment =~m{\S} ) {
		if ( ! defined $old_comment || $comment ne $old_comment ) {
			$x->SetNewValue( 'Comment', $comment );
			$changed++;
		}
	} elsif ( defined $old_comment ) {
		$x->SetNewValue( 'Comment' );
		$changed++;
	}

	if ( defined $caption && $caption =~m{\S} ) {
		if ( !defined $old_caption || $caption ne $old_caption ) {
			$x->SetNewValue( 'By-lineTitle', $caption );
			$changed++;
		}
	} elsif ( defined $old_caption ) {
		$x->SetNewValue( 'By-lineTitle' );
		$changed++;
	}

	if ( $changed ) {
		DEBUG( "update EXIF/JPEG/IPTCP info for $file" );
		unlink( "$file.bak" );
		rename($file,"$file.bak") or die $!;
		$x->WriteInfo( "$file.bak",$file );
		-s $file or do {
			warn "failed to write $file, restoring from backed up version...\n";
			unlink( $file );
			rename( "$file.bak",$file );
		};
	}
	1;
}


sub DEBUG {
	$DEBUG or return;
	print STDERR "DEBUG: @_\n";
}
